#
# Python bindings for TinySoundFont
# https://github.com/nwhitehead/tinysoundfont-pybind
#
# Copyright (C) 2024 Nathan Whitehead
#
# This code is licensed under the MIT license (see LICENSE for details)
#

from .._tinysoundfont import _midi_load_memory
from .._tinysoundfont import MidiMessageType
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional


@dataclass
class NoteOn:
    """Action that turns on a single note

    :param key: MIDI note number of note (0-127)
    :param velocity: Velocity of note, 0 means turn off (0-127)
    """

    key: int
    velocity: int = 0


@dataclass
class NoteOff:
    """Action that turns off a single note

    :param key: MIDI note number of note (0-127)
    """

    key: int


@dataclass
class ControlChange:
    """Action that changes a MIDI control value

    :param control: MIDI controller number (0-127)
    :param control_value: Value to change to (0-127)
    """

    control: int
    control_value: int


@dataclass
class ProgramChange:
    """Action that changes current preset program

    :param program: Preset number to use (0-127)
    """

    program: int


@dataclass
class PitchBend:
    """Action that bends pitch of current channel

    :param pitch_bend: How much to bend pitch (0-16383, 8192 means no bend)

    Range of pitch bend is set using RPN
    """

    pitch_bend: int = 8192


Action = NoteOn | NoteOff | ControlChange | ProgramChange | PitchBend


@dataclass
class Event:
    """
    A single event that can be sent to a :class:`Synth` object.

    :param action: Details of what event is
    :param t: Time when event is scheduled, in absolute seconds
    :param channel: Channel to use for event
    :param persistent: Whether to keep event in event list after playing

    """

    action: Action
    t: float = 0
    channel: int = 0
    persistent: bool = True


def event_from_dict(item: Dict) -> Optional[Event]:
    """Convert a dictionary with fields to an Event.

    :param item: Dictionary with fields `t` for time, `type` of type :class:`MidiEventType`, and other fields
    as needed for the event

    :returns: Converted input or `None` if the input could not be converted (unsupported type etc.)
    """
    t = item["t"]
    channel = item["channel"]
    persistent = "persistent" in item and item["persistent"] is not None
    match item["type"]:
        case MidiMessageType.NOTE_ON:
            return Event(
                NoteOn(item["key"], item["velocity"]),
                t=t,
                channel=channel,
                persistent=persistent,
            )
        case MidiMessageType.NOTE_OFF:
            return Event(
                NoteOff(item["key"]), t=t, channel=channel, persistent=persistent
            )
        case MidiMessageType.CONTROL_CHANGE:
            return Event(
                ControlChange(item["control"], item["control_value"]),
                t=t,
                channel=channel,
                persistent=persistent,
            )
        case MidiMessageType.PROGRAM_CHANGE:
            return Event(
                ProgramChange(item["program"]),
                t=t,
                channel=channel,
                persistent=persistent,
            )
        case MidiMessageType.PITCH_BEND:
            return Event(
                PitchBend(item["pitch_bend"]),
                t=t,
                channel=channel,
                persistent=persistent,
            )
        case MidiMessageType.KEY_PRESSURE:
            return None
        case MidiMessageType.CHANNEL_PRESSURE:
            return None


def load_memory(
    data: bytes,
    delta_time: float = 0,
    filter: Optional[Callable[[List[Event]], Optional[bool]]] = None,
    persistent: bool = True,
) -> List[Event]:
    """Load MIDI data and turn into list of events.

    :param data: MIDI data, in Standard MIDI format
    :param delta_time: Time offset to add to all events (default 0)
    :param filter: Optional function that takes in individual events and can
        modify them or filter them out (default None)
    :param persistent: Whether to keep events in queue after playing,
        allowing for seeking back to start or arbitrary positions after
        playback has started (default True)

    :returns: List of events from MIDI data, possibly filtered

    Filter function should take one input argument, the event, and modify it
    in place as needed. The function should return `None` or `False` to
    indicate to keep the modified event, or `True` to indicate that the
    event should be deleted.

    See also: :meth:`load`
    """
    midi_data = _midi_load_memory(data)
    events = []
    for item in midi_data:
        event = event_from_dict(item)
        if event is None:
            continue
        # Offset by time_delta
        event.t += delta_time
        event.persistent = persistent
        if filter is not None:
            if filter(event):
                event = None
        # Check for None return value in filtered events, don't add if None
        if event is not None:
            events.append(event)
    events.sort(key=lambda e: e.t)
    return events


def load(
    filename: str,
    delta_time: float = 0,
    filter: Optional[Callable[[List[Event]], Optional[bool]]] = None,
    persistent: bool = True,
) -> List[Event]:
    """Load MIDI file and turn into list of events.

    :param filename: Filename to load MIDI data from, in Standard MIDI
        format
    :param delta_time: Time offset to add to all events (default 0)
    :param filter: Optional function that takes in individual events and can
        modify them or filter them out (default None)
    :param persistent: Whether to keep events in queue after playing,
        allowing for seeking back to start or arbitrary positions after
        playback has started (default True)

    :returns: List of events from MIDI data, possibly filtered

    Filter function should take one input argument, the event, and modify it
    in place as needed. The function should return `None` or `False` to
    indicate to keep the modified event, or `True` to indicate that the event
    should be deleted.

    See also: :meth:`load_memory`
    """
    with open(filename, "rb") as fin:
        data = fin.read()
        return load_memory(
            data, delta_time=delta_time, filter=filter, persistent=persistent
        )




import tinysoundfont
import time

synth = tinysoundfont.Synth()
sfid = synth.sfload("FluidR3_GM.sf2")
synth.program_select(0, sfid, 0, 0, True)

seq = tinysoundfont.Sequencer(synth)
seq.midi_load("drum.mid")
synth.start(buffer_size=4096)

while not seq.is_empty():
    time.sleep(0.5)
time.sleep(2.0)



#
# Python bindings for TinySoundFont
# https://github.com/nwhitehead/tinysoundfont-pybind
#
# Copyright (C) 2024 Nathan Whitehead
#
# This code is licensed under the MIT license (see LICENSE for details)
#

from . import _tinysoundfont

from typing import Optional

MAX_CHANNELS = 16


class SoundFontException(Exception):
    """An exception raised from tinysoundfont"""

    pass


class Synth:
    """Create new synthesizer object to control sound generation.

    :param gain: scale factor for audio output, in relative dB (default 0.0)
    :param samplerate: output samplerate in Hz (default 44100)

    If you need to mix many simultaneous voices you may need to turn down the
    `gain` to avoid clipping. Some SoundFonts also require gain adjustment to
    avoid being too loud or too quiet.
    """

    def _get_soundfont(self, sfid):
        if sfid not in self.soundfonts:
            raise SoundFontException("Invalid SoundFont id")
        return self.soundfonts[sfid]

    def _get_sfid(self, chan):
        if chan not in self.channel:
            raise SoundFontException("Invalid channel (channel not assigned)")
        return self.channel[chan]

    def __init__(self, gain: float = 0, samplerate: int = 44100):
        self.p = None
        self.stream = None
        self.gain = gain
        self.samplerate = samplerate
        # soundfonts maps sfid numbers to SoundFont objects
        self.soundfonts = {}
        # Unique identifier creator for this Synth
        self.next_sfid = 0
        # Keep track of which SoundFont to use for different channels
        # Dictionary of channel -> sfid
        self.channel = {}
        # Function to call to perform actions during audio callback
        self.callback = None

    def sfload(
        self, filename_or_bytes: str | bytes, gain: float = 0.0, max_voices: int = 256
    ) -> int:
        """Load SoundFont and return its ID

        :param filename_or_bytes: either a filename containing sf2/sf3/sfo
            SoundFont data or bytes object
        :param gain: gain adjustment for this SoundFont, in relative dB (default
            0.0)
        :param max_voices: maximum number of simultaneous voices (default 256)

        :return: ID of SoundFont to be used by other methods such as
            :func:`program_select`

        When deciding on the value for `max_voices`, one note in a SoundFont may
        use more than one voice. Playing multiple notes also uses more voices.
        If more voices are required than are available, older voices will be cut
        off.

        See also: :meth:`program_select`, :meth:`sfpreset_name`,
        :meth:`sfunload`
        """
        soundfont = _tinysoundfont.SoundFont(filename_or_bytes)
        soundfont.set_output(
            _tinysoundfont.OutputMode.StereoInterleaved,
            self.samplerate,
            self.gain + gain,
        )
        soundfont.set_max_voices(max_voices)
        sfid = self.next_sfid
        self.next_sfid += 1
        self.soundfonts[sfid] = soundfont
        # Set any unassigned channels to use this SoundFont
        for chan in range(MAX_CHANNELS):
            if chan not in self.channel:
                self.channel[chan] = sfid
        return sfid

    def sfunload(self, sfid: int):
        """Unload a SoundFont and free memory it used.

        :param sfid: ID of SoundFont to unload, as returned by :func:`sfload`
        :type sfid: int

        :raises: `SoundFontException` if the SoundFont does not exist

        See also: :meth:`sfload`
        """
        _ = self._get_soundfont(sfid)
        del self.soundfonts[sfid]
        # Clear any channels that refers to this sfid
        self.channel = {
            chan: self.channel[chan]
            for chan in self.channel
            if self.channel[chan] != sfid
        }

    def program_select(
        self, chan: int, sfid: int, bank: int, preset: int, is_drums: bool = False
    ):
        """Select a program from a SoundFont for specific channel

        :param chan: Channel to affect (0-15)
        :param sfid: ID of SoundFont to use
        :param bank: Bank to set (0-127)
        :param preset: Which preset to use (0-127)
        :param is_drums: Whether to set channel to MIDI drum mode (default
            False)

        :raises: `SoundFontException` if the SoundFont does not exist
        :raises: `RuntimeError` if channel, bank, or preset is out of range
        :raises: `RuntimeError` if bank/preset does not match any instrument for
            the SoundFont

        Note that presets are numbered with 0-based indexing. MIDI user
        interfaces typically number presets 1-128.
        """
        soundfont = self._get_soundfont(sfid)
        self.channel[chan] = sfid
        soundfont.channel_set_bank(chan, bank)
        soundfont.channel_set_preset_number(chan, preset, is_drums)

    def program_unset(self, chan: int):
        """Set the preset of a MIDI channel to an unassigned state.

        :param chan: Channel to affect (0-15)

        :raises: `SoundFontException` if channel is out of range
        """
        if chan not in self.channel:
            raise SoundFontException("Invalid channel (channel not assigned)")
        del self.channel[chan]

    def program_change(self, chan: int, preset: int, is_drums: bool = False):
        """Select a program for a specific channel.

        :param chan: Channel to affect (0-15)
        :param preset: Which preset to use (0-127)
        :param is_drums: Whether to set channel to MIDI drum mode (default
            False)

        :raises: `SoundFontException` if channel is out of range or does not
            have a SoundFont loaded
        :raises: `RuntimeError` if bank/preset are out of range or do not match
            any instrument for the SoundFont in the channel

        Note that presets are numbered with 0-based indexing. MIDI user
        interfaces typically number presets 1-128.
        """
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_set_preset_number(chan, preset, is_drums)

    def program_info(self, chan: int) -> (int, int, int):
        """Get SoundFont id, bank, program number, and preset number of channel.

        :param chan: Channel to use (0-15)

        :raises: `SoundFontException` if channel is out of range or has no
            SoundFont loaded

        :return: Tuple containing `(sfid, bank, preset)` indicating SoundFont
            ID, bank number, and preset number
        """
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        bank = soundfont.channel_get_preset_bank(chan)
        preset = soundfont.channel_get_preset_number(chan)
        return (sfid, bank, preset)

    def sfpreset_name(self, sfid: int, bank: int, preset: int) -> Optional[str]:
        """Return name of a SoundFont preset.

        :param sfid: ID of SoundFont to use
        :param bank: Bank to use (0-127)
        :param preset: Which preset to retrieve (0-127)

        :raises: `SoundFontException` if channel is out of range or has no
            SoundFont loaded
        :raises: `RuntimeError` if bank/preset are out of range

        :return: Name of preset in SoundFont, or `None` if preset does not exist
            in SoundFont

        Note that presets are numbered with 0-based indexing. MIDI user
        interfaces typically number presets 1-128.
        """
        soundfont = self._get_soundfont(sfid)
        name = soundfont.get_preset_name(bank, preset)
        if name == "<None>":
            return None
        return name

    def noteon(self, chan: int, key: int, velocity: int) -> bool:
        """Play a note.

        :param chan: Channel to use (0-15)
        :param key: MIDI key to press (0-127), 60 is middle C
        :param velocity: Velocity of keypress (0-127), 0 means to turn off, 127
            is maximum

        :return: `True` if note was valid, `False` if note was outside of legal
            range or channel did not have instrument loaded
        """
        if key < 0 or key > 127:
            return False
        if velocity < 0 or velocity > 127:
            return False
        if chan not in self.channel:
            return False
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_note_on(chan, key, velocity / 127.0)
        return True

    def noteoff(self, chan: int, key: int):
        """Stop a note.

        :param chan: Channel to use (0-15)
        :param key: MIDI key to release (0-127), 60 is middle C

        :return: `True` if note was valid, `False` if note was outside of legal
            range or channel did not have instrument loaded

        It is valid to call `noteoff` on a note that never had `noteon`.
        """
        if key < 0 or key > 127:
            return False
        if chan not in self.channel:
            return False
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_note_off(chan, key)
        return True

    def notes_off(self, chan: Optional[int] = None):
        """Turn off all playing notes in all channels or one specific channel.

        :param chan: Channel to use (0-15) or None to indicate all channels

        Some instruments have long decays or may continue to produce sound after
        a NOTE_OFF event. If you need all sounds to stop playing use
        :meth:`sounds_off`.
        """
        if chan is None:
            for chan in range(16):
                self.control_change(chan, 123, 0)
        else:
            self.control_change(chan, 123, 0)

    def sounds_off(self, chan: Optional[int] = None):
        """Turn off all playing sounds in all channels or one specific channel.

        :param chan: Channel to use (0-15) or None to indicate all channels

        Some instruments have long decays or may continue to produce sound after
        a NOTE_OFF event. If you need all notes to stop playing and continue
        producing the decay, use :meth:`notes_off`.
        """
        if chan is None:
            for chan in range(16):
                self.control_change(chan, 120, 0)
        else:
            self.control_change(chan, 120, 0)

    def control_change(self, chan: int, controller: int, control_value: int):
        """Change control value for a specific channel.

        :param chan: Channel to use (0-15)
        :param controller: Controller to update, (0-127), meaning defined by
            MIDI 1.0 standard
        :param control_value: Value to use for update, (0-127)

        The interpretation of controller number is from the MIDI 1.0 standard.
        Supported controller values that have an effect are:

        .. include:: table_supported_controller.rstinc

        The supported Registered Parameter Names (RPN) are:

        .. include:: table_supported_rpn.rstinc

        .. include:: note_rpn.rstinc
        """
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_midi_control(chan, controller, control_value)

    def set_tuning(self, chan: int, tuning: float):
        """Set tuning for a channel.

        :param chan: Channel to affect (0-15)
        :param tuning: Tuning adjustment in semitones (default 0.0)
        """
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_set_tuning(chan, tuning)

    def pitchbend(self, chan: int, value: int):
        """Set pitch wheel position for a channel.

        :param chan: Channel to affect (0-15)
        :param value: Value from 0 to 16383 indicating pitch bend down to pitch
            bend up (default 8192, no pitch change)

        See also: :meth:`pitchbend_range`
        """
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_set_pitch_wheel(chan, value)

    def pitchbend_range(self, chan: int, semitones: float):
        """Set pitch bend range up and down for a channel.

        :param chan: Channel to affect (0-15)
        :param semitones: Pitch bend range up and down in semitones (default
            2.0)

        See also: :meth:`pitchbend`
        """
        sfid = self._get_sfid(chan)
        soundfont = self._get_soundfont(sfid)
        soundfont.channel_set_pitch_range(chan, semitones)

    def start(self, buffer_size: int = 1024, **kwargs):
        """Start audio playback in a separate thread.

        :param buffer_size: Number of samples to buffer or 0 for automatic
            sizing for low latency (default 1024)

        Extra keyword arguments will be passed to the `pyaudio` stream
        constructor. Useful arguments might include:

        * `output_device_index` -- index of output device to use, or `None` for
          default output device

        Note that depending on your platform `pyaudio` may recognize a large
        number of devices not all of which are suitable for audio playback. You
        may need to use `pyaudio` method `get_host_api_info_by_index()` to find
        details about the `pyaudio` devices and choose a suitable index.

        The separate audio playback thread will continue generating and playing
        samples until stopped with :meth:`stop`.

        The audio thread will not prevent the main thread from exiting. If you
        turn on notes and call :meth:`start`, your main thread will need to call
        :func:`time.sleep` to let time pass to be able to hear the notes
        playing. To schedule note events through time see :class:`Sequencer`.

        See also: :meth:`stop`
        """

        # Import pyaudio here so if this function is not used there is no dependency
        import pyaudio

        def callback(in_data, frame_count, time_info, status):
            buffer = self.generate(samples=frame_count)
            # PyAudio needs actual bytes, not just memoryview
            return (bytes(buffer), pyaudio.paContinue)

        self.p = pyaudio.PyAudio()
        self.stream = self.p.open(
            format=pyaudio.paFloat32,
            channels=2,
            rate=self.samplerate,
            output=True,
            stream_callback=callback,
            frames_per_buffer=buffer_size,
            **kwargs
        )

    def stop(self):
        """Stop audio playback thread.

        See also: :meth:`start`
        """
        if self.p is not None and self.stream is not None:
            self.stream.close()
            self.p.terminate()

    def generate(self, samples: int, buffer: Optional[memoryview] = None) -> memoryview:
        """Generate fixed number of output samples.

        :param samples: Number of samples to generate
        :param buffer: Existing buffer to fill, or `None` to allocate new buffer

        :returns: View into buffer with samples filled in stereo float32 format.

        This method fills in a fixed number of output samples in the output
        buffer given (or creates a new buffer if none is given). The sequencer
        callback is called as needed to trigger MIDI events at the correct
        sample location.
        """
        CHANNELS = 2
        SIZEOF_FLOAT_IN_BYTES = 4
        if buffer is None:
            # Wrap with `memoryview` so slicing is references inside the buffer, not copies
            buffer = memoryview(bytearray(samples * CHANNELS * SIZEOF_FLOAT_IN_BYTES))
        generated = 0
        while generated < samples:
            delta = (samples - generated) / self.samplerate
            # Call all the relevant callbacks, which each may shorten delta
            # Each callback does any actions it needs to do that are currently scheduled, then returns how much delta can advance
            if self.callback is not None:
                delta = min(self.callback(delta), delta)
            # Compute actual frame count to render based on return value in seconds (round up to keep making progress in each iteration)
            actual_frame_count = int(delta * self.samplerate + 0.999)
            # Index into buffer at frame boundaries
            pos = generated * CHANNELS * SIZEOF_FLOAT_IN_BYTES
            sz_bytes = actual_frame_count * CHANNELS * SIZEOF_FLOAT_IN_BYTES
            self.generate_simple(actual_frame_count, buffer=buffer[pos : pos + sz_bytes])
            generated += actual_frame_count
        return buffer

    def generate_simple(self, samples: int, buffer: Optional[memoryview] = None) -> memoryview:
        """Generate fixed number of output samples, ignoring sequenced events.

        :param samples: Number of samples to generate
        :param buffer: Existing buffer to fill, or `None` to allocate new buffer

        :returns: View into buffer with samples filled in stereo float32 format.

        This method fills in a fixed number of output samples in the output
        buffer given (or creates a new buffer if none is given). The sequencer
        is not called from this method so no new events are ever triggered by
        this method.

        See also: :meth:`generate`
        """
        CHANNELS = 2
        SIZEOF_FLOAT_IN_BYTES = 4
        if buffer is None:
            buffer = memoryview(bytearray(samples * CHANNELS * SIZEOF_FLOAT_IN_BYTES))
        mix = False
        for soundfont in self.soundfonts.values():
            soundfont.render(buffer, mix)
            # After first render turn on mix to mix together all sounds
            mix = True
        return buffer




import tinysoundfont
import time

synth = tinysoundfont.Synth()
sfid = synth.sfload("FluidR3_GM.sf2")

seq = tinysoundfont.Sequencer(synth)
seq.midi_load("1080-c01.mid")

# Larger buffer because latency is not important
synth.start(buffer_size=4096)

while not seq.is_empty():
    time.sleep(0.5)




import numpy as np
import pydoc
import scipy.io.wavfile
import tempfile
import time
import zlib

import tinysoundfont

PAN_CONTROL = 10

def test_help():
    # Just make sure there is some text for `help(tinysoundfont)`
    helptext = pydoc.render_doc(tinysoundfont, "%s")
    assert len(helptext.split("\n")) > 10


def test_load():
    s = tinysoundfont.Synth()
    sfid = s.sfload("test/florestan-piano.sf2")
    assert s.sfpreset_name(sfid, 0, 0) == "Piano"
    with open("test/florestan-piano.sf2", "rb") as f:
        mem = f.read()
        assert mem[:4] == b"RIFF"
        assert len(mem) == 187548
        sfid2 = s.sfload(mem)
        assert s.sfpreset_name(sfid2, 0, 0) == "Piano"


def test_bytes():
    s = tinysoundfont.Synth(gain=-14)
    sfid = s.sfload("test/florestan-piano.sf2")
    s.program_select(0, sfid, 0, 0)
    s.noteon(0, 48, 100)
    # Create 1 second buffer
    buffer = s.generate(44100)
    # buffer now contains audio data
    # Check first and last value
    assert buffer[:4] == b"\x99\xfa\xac\xba"
    assert buffer[-4:] == b"\xc6z\x97;"


def test_wav():
    s = tinysoundfont.Synth(gain=-14)
    sfid = s.sfload("test/florestan-piano.sf2")
    s.program_select(0, sfid, 0, 0)
    s.program_select(1, sfid, 0, 0)
    s.program_select(2, sfid, 0, 0)

    buffer = bytearray()
    buffer.extend(s.generate(44100))

    s.noteon(0, 48, 100)
    s.noteon(1, 52, 100)
    s.noteon(2, 55, 100)

    buffer.extend(s.generate(44100))

    # 1 second pitch bend

    s.control_change(0, PAN_CONTROL, 0)
    s.control_change(1, PAN_CONTROL, 64)
    s.control_change(2, PAN_CONTROL, 127)
    s.pitchbend(0, 16737)
    s.pitchbend(1, 16737)
    s.pitchbend(2, 16737)

    buffer.extend(s.generate(44100))

    # 1 second fadeout

    s.noteoff(0, 48)
    s.noteoff(1, 52)
    s.noteoff(2, 55)

    buffer.extend(s.generate(44100))

    # Convert bytearray value into numpy (samples, 2) matrix in float32 format
    output = np.reshape(np.frombuffer(buffer, dtype=np.float32), (-1, 2))

    # Write to WAV file
    with tempfile.NamedTemporaryFile(mode="wb", suffix=".wav") as wavfile:
        scipy.io.wavfile.write(wavfile.name, 44100, output)
        with open(wavfile.name, "rb") as wavfileread:
            contents = wavfileread.read()
            # Check that WAV file has exactly right size
            assert len(contents) == 1411258
            # Check that WAV has right CRC32 (golden value)
            assert zlib.crc32(contents) == 3997000505


def test_start():
    s = tinysoundfont.Synth(gain=-14)
    sfid = s.sfload("test/florestan-piano.sf2")
    s.program_select(0, sfid, 0, 0)
    s.program_select(1, sfid, 0, 0)
    s.program_select(2, sfid, 0, 0)
    s.pitchbend_range(0, 12.0)
    s.pitchbend_range(1, 12.0)
    s.pitchbend_range(2, 12.0)
    s.start(buffer_size=2048)

    time.sleep(1.0)

    s.noteon(0, 48, 100)
    s.noteon(1, 52, 100)
    s.noteon(2, 55, 100)

    time.sleep(1.0)

    s.pitchbend(0, 0)
    s.pitchbend(1, 8192)
    s.pitchbend(2, 0)
    s.control_change(0, PAN_CONTROL, 0)
    s.control_change(1, PAN_CONTROL, 64)
    s.control_change(2, PAN_CONTROL, 127)

    time.sleep(1.0)

    s.noteoff(0, 48)
    s.noteoff(1, 52)
    s.noteoff(2, 55)

    time.sleep(1.0)

    s.stop()




import tinysoundfont
import time

synth = tinysoundfont.Synth()
sfid = synth.sfload("florestan-subset.sfo")
synth.program_select(0, sfid, 0, 12) # Marimba
synth.program_select(1, sfid, 0, 45) # PizzicatoStr
synth.program_select(2, sfid, 0, 24) # Nylon-str.Gt
synth.start()

time.sleep(0.5)
synth.noteon(0, 60, 100)
synth.noteon(0, 67, 100)
time.sleep(0.5)
synth.noteon(1, 60, 100)
time.sleep(0.5)
synth.noteon(2, 70, 75)
time.sleep(0.5 / 8)
synth.noteoff(2, 70)
synth.noteon(2, 67, 75)

time.sleep(1.0)
synth.noteoff(0, 60)
synth.noteoff(0, 67)
synth.noteoff(1, 60)
synth.noteoff(2, 67)

time.sleep(1.0)




import tinysoundfont
import time

synth = tinysoundfont.Synth()
sfid = synth.sfload("florestan-subset.sfo")

def filter_program_change(event):
    """Make all program changes go to preset 19"""
    match event.action:
        case tinysoundfont.midi.ProgramChange(_program):
            event.action.program = 19 # Church organ

seq = tinysoundfont.Sequencer(synth)
seq.midi_load("1080-c01.mid", filter=filter_program_change)

synth.start(buffer_size=4096)
while not seq.is_empty():
    time.sleep(0.5)



import argparse
import sys
import time

from .synth import Synth
from .sequencer import Sequencer


def endswith_any(value, suffixes):
    for suffix in suffixes:
        if value.endswith(suffix):
            return True
    return False


def is_midi(filename):
    return endswith_any(filename, [".mid", ".midi", ".MID", ".MIDI"])


def is_soundfont(filename):
    return endswith_any(filename, [".sf2", ".SF2", ".sf3", ".SF3", ".sfo", ".SFO"])


def main():
    parser = argparse.ArgumentParser(
        prog="tinysoundfont-tool",
        description="Command line tool for examining SoundFonts and MIDI files",
    )
    parser.add_argument(
        "--play",
        action="store_true",
        help="Play MIDI file (requires both MIDI and SoundFont files)",
    )
    parser.add_argument("--test", action="store_true", help="Play test SoundFont file")
    parser.add_argument(
        "--info", action="store_true", help="Show information about SoundFont file"
    )
    parser.add_argument(
        "--key",
        type=int,
        action="append",
        default=[],
        help="MIDI key number to play for test (0-127)",
    )
    parser.add_argument(
        "--bank",
        type=int,
        action="append",
        default=[],
        help="MIDI bank number to use for playing test (0-127)",
    )
    parser.add_argument(
        "--preset",
        type=int,
        action="append",
        default=[],
        help="MIDI preset number to use for playing test (0-127)",
    )
    parser.add_argument(
        "--velocity",
        type=int,
        action="append",
        default=[],
        help="MIDI velocity to use for playing test (0-127)",
    )
    parser.add_argument(
        "--drum", action="store_true", help="Set drum mode for play test"
    )
    parser.add_argument(
        "--samplerate",
        type=int,
        default=44100,
        help="Set synthesizer samplerate for playback",
    )
    parser.add_argument(
        "--gain",
        type=float,
        default=0.0,
        help="Set synthesizer global gain for playback",
    )
    parser.add_argument(
        "--buffer_size",
        type=int,
        default=2048,
        help="Set pyaudio buffer size for playback, or 0 for automatic sizing for lowest latency",
    )
    parser.add_argument(
        "filename",
        nargs="+",
        help="SoundFont or MIDI files to use; playing MIDI files requires both MIDI file and SoundFont file",
    )
    args = parser.parse_args()

    midi_filename = None
    soundfont_filename = None
    for filename in args.filename:
        if is_midi(filename):
            midi_filename = filename
        if is_soundfont(filename):
            soundfont_filename = filename

    if args.info:
        if soundfont_filename is None:
            print(
                "No SoundFont file found, a SoundFont file is required for MIDI playback"
            )
            return -2
        print(f"Info for SoundFont {soundfont_filename}")
        synth = Synth(samplerate=args.samplerate, gain=args.gain)
        sfid = synth.sfload(soundfont_filename)
        for bank in range(127):
            for preset in range(127):
                name = synth.sfpreset_name(sfid, bank, preset)
                if name is not None:
                    print(f"{bank} - {preset} : {name}")
        return 0

    if args.test:
        if soundfont_filename is None:
            print(
                "No SoundFont file found, a SoundFont file is required for MIDI playback"
            )
            return -2
        print(f"Testing SoundFont {soundfont_filename}")
        synth = Synth(samplerate=args.samplerate, gain=args.gain)
        sfid = synth.sfload(soundfont_filename)
        if len(args.bank) == 0:
            args.bank.append(0)
        if len(args.preset) == 0:
            args.preset.append(0)
        if len(args.key) == 0:
            args.key.append(60)
        if len(args.velocity) == 0:
            args.velocity.append(100)
        for i in range(16):
            synth.program_select(
                i,
                sfid,
                args.bank[i % len(args.bank)],
                args.preset[i % len(args.preset)],
                is_drums=args.drum,
            )
        for i in range(len(args.key)):
            synth.noteon(
                i, args.key[i % len(args.key)], args.velocity[i % len(args.velocity)]
            )
        synth.start(buffer_size=args.buffer_size)
        time.sleep(3)
        synth.notes_off()
        time.sleep(1)

    if args.play:
        if midi_filename is None:
            print(
                "No MIDI file found, a MIDI file and SoundFont file are required for MIDI playback"
            )
            return -1
        if soundfont_filename is None:
            print(
                "No SoundFont file found, a SoundFont file is required for MIDI playback"
            )
            return -2

        # Let's try to play a MIDI file
        synth = Synth(samplerate=args.samplerate, gain=args.gain)
        sfid = synth.sfload(soundfont_filename)
        seq = Sequencer(synth)
        for i in range(16):
            synth.program_change(i, 0, i == 10)
        seq.midi_load(midi_filename)
        synth.start(buffer_size=args.buffer_size)
        print(f"Playing {midi_filename} with SoundFont {soundfont_filename}")
        while not seq.is_empty():
            time.sleep(0.5)
        time.sleep(1)
        return 0

        return 0

    print("No action to perform, need either --test, --play, or --info")
    return -3


sys.exit(main())



import pytest
import tinysoundfont
import time


def test_0():
    assert int(tinysoundfont.midi.MidiMessageType.NOTE_ON) == 0x90

    with open("test/1080-c01.mid", "rb") as fin:
        contents = fin.read()
        data = tinysoundfont.midi._midi_load_memory(contents)
        assert len(data) == 1852
        assert data[0] == {
            "t": 0.0,
            "type": tinysoundfont.midi.MidiMessageType.SET_TEMPO,
            "channel": 7,
            "bpm": 115.00002875000719,
        }
        assert data[4] == {
            "t": 0.0,
            "type": tinysoundfont.midi.MidiMessageType.PROGRAM_CHANGE,
            "channel": 1,
            "program": 40,
            "bpm": 115.00002875000719,
        }
        assert data[7] == {
            "t": 0.0,
            "type": tinysoundfont.midi.MidiMessageType.NOTE_ON,
            "channel": 1,
            "key": 62,
            "velocity": 100,
            "bpm": 115.00002875000719,
        }

    synth = tinysoundfont.Synth(samplerate=22050, gain=-3.0)
    # Try with buffer size large enough that listener will hear jitter if notes must start/end on buffer boundaries
    synth.start(buffer_size=4096)

    sfid = synth.sfload("test/florestan-piano.sf2", gain=-12.0)
    assert sfid == 0
    sfid2 = synth.sfload("test/florestan-subset.sfo", gain=-1.0)
    assert sfid2 == 1
    name = synth.sfpreset_name(sfid, 0, 0)
    assert name == "Piano"
    name = synth.sfpreset_name(sfid2, 0, 2)
    assert name == "Piano"
    synth.program_select(0, sfid, 0, 0)
    info = synth.program_info(0)
    assert info == (0, 0, 0)
    synth.program_select(1, sfid2, 0, 2)

    synth.noteon(0, 48, 100)
    synth.noteon(1, 60, 100)
    synth.noteon(1, 64, 100)
    synth.noteon(1, 67, 100)
    time.sleep(0.5)
    # Tune channel 0 up half a semitone
    synth.set_tuning(0, 0.5)
    time.sleep(0.5)
    synth.noteoff(0, 48)
    synth.noteoff(1, 60)
    synth.noteoff(1, 64)
    synth.noteoff(1, 67)
    time.sleep(1.0)

    synth.sfunload(sfid)
    synth.sfunload(sfid2)
    sfid2 = synth.sfload("test/florestan-subset.sfo", gain=-1.0)
    for bank in range(127):
        for i in range(127):
            name = synth.sfpreset_name(sfid2, bank, i)
            if name:
                print(bank, i, name)

    def filter_program_change(event):
        """Make all program changes go to preset 40 (violin)"""
        match event.action:
            case tinysoundfont.midi.ProgramChange(_program):
                event.program = 40
    seq = tinysoundfont.Sequencer(synth)
    seq.midi_load("test/1080-c01.mid", filter=filter_program_change)
    time.sleep(10)

    with pytest.raises(Exception):
        synth.sfunload(sfid)
    synth.sfunload(sfid2)
    synth.stop()

def test_midi_generate():
    import numpy as np

    synth = tinysoundfont.Synth()
    _sfid = synth.sfload("test/florestan-subset.sfo")

    seq = tinysoundfont.Sequencer(synth)
    seq.midi_load("test/1080-c01.mid")

    # Generate 1 second buffer
    buffer = synth.generate(44100)
    block = np.frombuffer(bytes(buffer), dtype=np.float32)
    assert block.min() < block.max()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import importlib.metadata
import packaging

project = 'tinysoundfont'
copyright = '2024, Nathan Whitehead'
author = 'Nathan Whitehead'

# The short X.Y version
project_version = importlib.metadata.version(project)
release = project_version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.githubpages',
    'sphinx_autodoc_typehints',
    'sphinx_tabs.tabs',
]
sphinx_tabs_disable_tab_closing = True
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'




import tinysoundfont
import time

synth = tinysoundfont.Synth()
sfid = synth.sfload("florestan-piano.sf2")
synth.program_select(0, sfid, 0, 0)
synth.start()

time.sleep(0.5)

synth.noteon(0, 48, 100)
synth.noteon(0, 52, 100)
synth.noteon(0, 55, 100)

time.sleep(0.5)

synth.noteoff(0, 48)
synth.noteoff(0, 52)
synth.noteoff(0, 55)

time.sleep(0.5)



#
# Python bindings for TinySoundFont
# https://github.com/nwhitehead/tinysoundfont-pybind
#
# Copyright (C) 2024 Nathan Whitehead
#
# This code is licensed under the MIT license (see LICENSE for details)
#

from collections import deque
from typing import List
from .synth import Synth
from .midi import (
    load,
    Event,
    NoteOn,
    NoteOff,
    ControlChange,
    ProgramChange,
    PitchBend,
)

DRUM_CHANNEL = 9

class Sequencer:
    """A Sequencer schedules MIDI events over time.

    :param synth: The synthesizer object to send events to.
    """

    def __init__(self, synth: Synth):
        self.synth = synth
        self.time = 0.0
        self.paused = False
        # events stores events for future, ordered by time
        # Queue has items (t, event)
        self.events = deque()

        def seq_callback(delta: float) -> float:
            if self.paused:
                return delta
            return self.process(delta)

        self.synth.callback = seq_callback

    def add(self, events: List[Event]):
        """Add a list of MIDI events to queue for sending.

        :param events: List of MIDI events

        See :func:`midi.load` for generating the list of events.
        See :func:`midi_load` for directly loading a MIDI file.
        """
        self.events.extend(events)

    def midi_load(self, filename: str, **kwargs):
        """Load MIDI file and schedule events.

        :param filename: Filename to load MIDI data from, in Standard MIDI
            format

        Any additional keyword arguments are passed to :func:`midi.load`. See
        :func:`midi.load` for documentation on additional keyword arguments.
        """
        events = load(filename, **kwargs)
        self.add(events)

    def get_time(self) -> float:
        """Get current playing time of sequencer.

        :return: current playing time of sequencer in seconds of absolute time since sequencer started
        """
        return self.time

    def set_time(self, time: float):
        """Set current playing time of sequencer.

        :param time: New absolute time in seconds

        Note that if previously scheduled events did not have `persistent` set
        then the events will no longer exist and will not play again.

        Note that MIDI expects events to happen in sequence. So if there is a
        program change as an event and you set the time to before that event,
        you might get the new instrument at a position where the previous
        instrument should have played. In general seeking back to the start of
        the MIDI events should work. Other random seeking may have unintended
        effects.

        To avoid stuck notes, this method turns off all keypresses using
        :meth:`notes_off`. It does not stop all sounds so playing notes may
        still have time to decay. If needed you can call :meth:`sounds_off`
        to stop all playing sounds immediately.
        """
        self.time = time
        self.notes_off()

    def pause(self, pause_value=True):
        """Pause or unpause playback.

        When playback is paused, time does not advance. No new events will be
        sent to the :class:`Synth` object connected to the sequencer. If notes
        are playing, they will continue to play.

        To prevent notes from continuing to play during pause, this method calls
        :meth:`notes_off`. You may want to call :meth:`sound_off` to stop all
        sound playing immediately if needed.
        """
        self.paused = pause_value
        self.notes_off()

    def notes_off(self):
        """Send NOTE_OFF for all currently playing notes of the Synth object."""
        self.synth.notes_off()

    def sounds_off(self):
        """Turn off all sounds of the Synth object."""
        self.synth.sounds_off()

    def is_empty(self) -> bool:
        """Return True if there are no more events scheduled.

        :return: True if there are no more events scheduled for this sequencer
            (song is over)

        Songs often end on a final NOTEOFF event that may have a decay time. It
        is often good to wait some amount of time before looping or scheduling a
        new song.
        """
        if len(self.events) == 0:
            return True
        if self.events[-1].t < self.time:
            return True
        return False

    def send(self, event: Event):
        """Send a single MIDI event to the synth object now, ignoring any time information.

        :param event: Single event to send

        If the event tries to select a preset that does not exist no error is raised.

        """
        synth = self.synth
        channel = event.channel
        match event.action:
            case NoteOn(key, velocity):
                synth.noteon(channel, key, velocity)
            case NoteOff(key):
                synth.noteoff(channel, key)
            case ControlChange(control, control_value):
                synth.control_change(channel, control, control_value)
            case ProgramChange(program):
                try:
                    synth.program_change(channel, program, channel == DRUM_CHANNEL)
                except Exception:
                    pass
            case PitchBend(pitch_bend):
                synth.pitchbend(channel, pitch_bend)

    def process(self, delta: float) -> float:
        """Advance time and send any events that need to be sent to the Synth.

        :param delta: How many seconds to advance time
        :returns: How far time was actually advanced (may be smaller than `delta`)

        """
        # Need to send all events with time <= self.time if not persistent
        # Persistent events only send events with time == self.time
        pos = 0
        while pos < len(self.events):
            # Look at earliest event in deque
            event = self.events[pos]
            t = event.t
            if t < self.time:
                # Send it if it is not persistent, otherwise ignore
                if event.persistent:
                    pos += 1
                else:
                    self.events.popleft()
                    self.send(event)
                continue
            if t > self.time:
                # Don't do event yet or remove it, need to wait until its time
                true_delta = min(delta, t - self.time)
                self.time += true_delta
                return true_delta
            # If we get here, t == self.time
            # Head event needs to be done
            if event.persistent:
                pos += 1
            else:
                self.events.popleft()
                # Don't change pos here, the popleft shifts pos to be next event
            # Now do it
            self.send(event)
        # If we get here that means events are done
        # Advance full time
        self.time += delta
        return delta
